#include <atmel_start.h>

#define wait_ready() \
    while (NVMCTRL->INTFLAG.bit.READY == 0) \
        ;

#define FLASH_ROW_SIZE 256

#define SECOND_APP_START_ADDRESS 0x2000

void flash_erase_row(uint32_t *dst)
{
    wait_ready();
    NVMCTRL->STATUS.reg = NVMCTRL_STATUS_MASK;

    // Execute "ER" Erase Row
    NVMCTRL->ADDR.reg = (uint32_t)dst / 2;
    NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_ER;
    wait_ready();
}

void flash_write_words(uint32_t *dst, uint32_t *src, uint32_t n_words)
{
    // Set automatic page write
    NVMCTRL->CTRLB.bit.MANW = 0;

    while (n_words > 0) {
        uint32_t len = (FLASH_PAGE_SIZE >> 2) < n_words ? (FLASH_PAGE_SIZE >> 2) : n_words;
        n_words -= len;

        // Execute "PBC" Page Buffer Clear
        NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_PBC;
        wait_ready();

        // make sure there are no other memory writes here
        // otherwise we get lock-ups

        while (len--)
            *dst++ = *src++;

        // Execute "WP" Write Page
        NVMCTRL->CTRLA.reg = NVMCTRL_CTRLA_CMDEX_KEY | NVMCTRL_CTRLA_CMD_WP;
        wait_ready();
    }
}

void flash_write_row(uint32_t *dst, uint32_t *src)
{
    bool src_different = false;
    for (int i = 0; i < FLASH_ROW_SIZE / 4; ++i) {
        if (src[i] != dst[i]) {
            src_different = true;
            break;
        }
    }

    if (src_different) {
            flash_erase_row(dst);
            flash_write_words(dst, src, FLASH_ROW_SIZE / 4);
    }
}

void flash_copy(uint32_t *dst, uint32_t *src, uint32_t size)
{
        uint32_t n_rows;
        n_rows = size / FLASH_ROW_SIZE;
        while (n_rows--)
        {
                flash_write_row(dst, src);
                dst += FLASH_ROW_SIZE / 4;
                src += FLASH_ROW_SIZE / 4;
        }
}

void jump_to_secondary()
{
        uint32_t app_start_address;
        app_start_address = *(uint32_t *)(SECOND_APP_START_ADDRESS + 4);

        /* Rebase the Stack Pointer */
        __set_MSP(*(uint32_t *)SECOND_APP_START_ADDRESS);

        /* Rebase the vector table base address */
        SCB->VTOR = ((uint32_t)SECOND_APP_START_ADDRESS & SCB_VTOR_TBLOFF_Msk);

        /* Jump to application Reset Handler in the application */
        asm("bx %0" ::"r"(app_start_address));
}

int comp_sysv_checksum(uint8_t *p_start, uint8_t *p_end)
{
        uint32_t result = 0;
        uint16_t result16 = 0;
        uint8_t *p = p_start;
        while (p != p_end)
                result += *(p++);
        result16 = (result & 0xffff) + ((result & 0xffffffff) >> 16);
        return (result16 & 0xffff) + (result16 >> 16);
}

int main(void)
{
        flash_copy((uint32_t*)0x2000, (uint32_t*)0x22000, 0x1e000);

        jump_to_secondary();
}
